/*
    $Id: callbacks.c,v 1.7 2001/12/08 03:48:44 root Exp root $

    k164, telephone call logger daemon
    Copyright (C) 2001  James Cameron (quozl@us.netrek.org)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <ctype.h>
#include <termios.h>
#include <unistd.h>
#include <string.h>
#include <glib.h>
#include <errno.h>
#include <fcntl.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"

int fd;			/* file descriptor for serial port	*/
guint in;		/* GTK input stream identifier		*/
char *device = NULL;	/* serial port device name		*/
#define MAXTIME 80	/* maximum width of timestamp		*/
#define MAXDATA 128	/* maximum width of DTMF data		*/

#define PRIVACY 4	/* number of dial digits to show	*/

int state = 0;		/* data stream parsing state		*/

GtkCList *clist;	/* the main window's columned list	*/
guint which;		/* current row within list being added	*/

#define MAXCOLUMNS 4	/* number of columns to populate	*/

guint context;			/* status bar context		*/
GtkStatusbar *statusbar;	/* status bar widget pointer	*/
GtkProgressBar *progressbar;

time_t t0, t1;		/* date and time off and on hook	*/

struct _Settings {
  guint log_to_file;
  guint log_to_stdout;
  guint comments;
  guint flag;
  guint privacy;
};

typedef struct _Settings Settings;

Settings *settings = NULL;

/* menu stuff */

void menu_hide(char *name)
{
  GtkWidget *widget = lookup_widget(GTK_WIDGET(clist), name);
  gtk_widget_hide(widget);
}

void menu_set_sensitive(char *name, int x)
{
  GtkWidget *widget = lookup_widget(GTK_WIDGET(clist), name);
  gtk_widget_set_sensitive(widget, x);
}

void menu_device_set_sensitive(int x)
{
  menu_set_sensitive("test", x);
  menu_set_sensitive("download_only", x);
  menu_set_sensitive("download_and_clear", x);
  menu_set_sensitive("set_date_and_time", x);
  menu_set_sensitive("change_password", x);
}

/* message window stuff */

gboolean
on_message_delete_event                (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  gtk_widget_destroy (gtk_widget_get_toplevel (GTK_WIDGET(widget)));
  return FALSE;
}


void
on_message_ok_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_destroy (gtk_widget_get_toplevel (GTK_WIDGET(button)));
}


static GtkWidget*
create_message ()
{
  GtkWidget *message;
  GtkWidget *dialog_vbox1;
  GtkWidget *message_label;
  GtkWidget *dialog_action_area1;
  GtkWidget *hbuttonbox2;
  GtkWidget *message_ok;

  message = gtk_dialog_new ();
  gtk_widget_set_name (message, "message");
  gtk_object_set_data (GTK_OBJECT (message), "message", message);
  GTK_WINDOW (message)->type = GTK_WINDOW_DIALOG;
  gtk_window_set_title (GTK_WINDOW (message), "unset title");
  gtk_window_set_position (GTK_WINDOW (message), GTK_WIN_POS_MOUSE);
  gtk_window_set_policy (GTK_WINDOW (message), TRUE, TRUE, TRUE);
  gtk_window_set_wmclass (GTK_WINDOW (message), "k164message", "K164message");

  dialog_vbox1 = GTK_DIALOG (message)->vbox;
  gtk_widget_set_name (dialog_vbox1, "dialog_vbox1");
  gtk_object_set_data (GTK_OBJECT (message), "dialog_vbox1", dialog_vbox1);
  gtk_widget_show (dialog_vbox1);

  message_label = gtk_label_new ("unset label");
  gtk_widget_set_name (message_label, "message_label");
  gtk_object_set_data (GTK_OBJECT (message), "message_label", message_label);
  gtk_widget_show (message_label);
  gtk_box_pack_start (GTK_BOX (dialog_vbox1), message_label, FALSE, FALSE, 0);
  gtk_misc_set_padding (GTK_MISC (message_label), 8, 8);

  dialog_action_area1 = GTK_DIALOG (message)->action_area;
  gtk_widget_set_name (dialog_action_area1, "dialog_action_area1");
  gtk_object_set_data (GTK_OBJECT (message), "dialog_action_area1", dialog_action_area1);
  gtk_widget_show (dialog_action_area1);
  gtk_container_set_border_width (GTK_CONTAINER (dialog_action_area1), 10);

  hbuttonbox2 = gtk_hbutton_box_new ();
  gtk_widget_set_name (hbuttonbox2, "hbuttonbox2");
  gtk_object_set_data (GTK_OBJECT (message), "hbuttonbox2", hbuttonbox2);
  gtk_widget_show (hbuttonbox2);
  gtk_box_pack_start (GTK_BOX (dialog_action_area1), hbuttonbox2, TRUE, TRUE, 0);

  message_ok = gtk_button_new_with_label ("OK");
  gtk_widget_set_name (message_ok, "message_ok");
  gtk_object_set_data (GTK_OBJECT (message), "message_ok", message_ok);
  gtk_widget_show (message_ok);
  gtk_container_add (GTK_CONTAINER (hbuttonbox2), message_ok);
  GTK_WIDGET_SET_FLAGS (message_ok, GTK_CAN_DEFAULT);
  gtk_widget_grab_focus (message_ok);
  gtk_widget_grab_default (message_ok);

  gtk_signal_connect (GTK_OBJECT (message), "delete_event",
                      GTK_SIGNAL_FUNC (on_message_delete_event),
                      NULL);
  gtk_signal_connect (GTK_OBJECT (message_ok), "clicked",
                      GTK_SIGNAL_FUNC (on_message_ok_clicked),
                      NULL);

  return message;
}

/* display a message window, which self deletes when user hits ok */
static GtkWidget *message (char *title, char *text)
{
  GtkWidget *widget, *label;

  widget = create_message();
  gtk_window_set_title(GTK_WINDOW (widget), title);
  label = lookup_widget(widget, "message_label");
  gtk_label_set(GTK_LABEL(label), text);
  gtk_widget_show(widget);
  return widget;
}


/* status bar stuff */

/* purge a specific message from the status bar stack */
static int statusbar_purge(guint *data)
{
  gtk_statusbar_remove(statusbar, context, *data);
  g_free(data);
  return FALSE;
}

/* push a message onto the status bar stack */
static guint statusbar_push(gchar *message)
{
  return gtk_statusbar_push(statusbar, context, message);
}

/* pop a specific message from the status bar stack */
/*
static void statusbar_pop(guint *message)
{
  if (*message != 0) {
    gtk_statusbar_remove(statusbar, context, *message);
    *message = 0;
  }
}
*/

/* push a timed message onto the status bar */
static void statusbar_timed(gchar *message, guint milliseconds)
{
  guint *reference = g_new(guint, 1);
  *reference = statusbar_push(message);
  gtk_timeout_add(milliseconds, (GtkFunction) statusbar_purge, reference);
}

/* initialise the status bar pointers */
void statusbar_init(GtkWidget *widget)
{
  statusbar = GTK_STATUSBAR(lookup_widget(widget, "statusbar"));
  progressbar = GTK_PROGRESS_BAR(lookup_widget(widget, "progressbar"));
  context = gtk_statusbar_get_context_id (statusbar, "reset");
}


/* entry list stuff */

/* append a dialled number to the current entry */
static void entry_append(int byte)
{
  gchar *buffer;
  gtk_clist_get_text(clist, which, 1, &buffer);
  if (settings->privacy && strlen(buffer) >= PRIVACY) return;
  buffer = g_strdup_printf("%s%c", buffer, byte);
  gtk_clist_set_text(clist, which, 1, buffer);
  g_free(buffer);
}

static char *to_elapsed(int seconds)
{
  int minutes = seconds / 60;
  return g_strdup_printf("%02d:%02d", minutes, seconds % 60);
}

/* update the elapsed time */
static void entry_update(int seconds)
{
  gchar *buffer = to_elapsed(seconds);
  gtk_clist_set_text(clist, which, 2, buffer);
  g_free(buffer);
}

guint entry_timeout_id;

static int entry_timeout(gpointer ignored)
{
  entry_update((int) (time(NULL) - t0));
  return TRUE;
}

/* begin an entry, called when off hook detected */
static void entry_begin(char *stamp)
{
  char *columns[MAXCOLUMNS];

  columns[0] = stamp+10;
  columns[1] = "";
  columns[2] = "";
  columns[3] = "";

  which = gtk_clist_append(clist, columns);
  gtk_clist_moveto(clist, which, 0, 1.0, 0.0);
  gtk_clist_select_row(clist, which, 0);
  entry_timeout_id = gtk_timeout_add(500, (GtkFunction) entry_timeout, NULL);
}

/* close an entry, called when on hook detected */
static void entry_close(int seconds, char *stamp)
{
  gtk_timeout_remove(entry_timeout_id);
  entry_update(seconds);
  gtk_clist_set_text(clist, which, 3, stamp+10);
  gtk_clist_unselect_row(clist, which, 0);
}

/* make a generic entry event, such as for a reset */
static void entry_generic(char *c1, char *c2, char *c3, char *c4)
{
  char *columns[MAXCOLUMNS];

  columns[0] = c1;
  columns[1] = c2;
  columns[2] = c3;
  columns[3] = c4;

  which = gtk_clist_append(clist, columns);
  gtk_clist_moveto(clist, which, 0, 1.0, 0.0);
}


/* dual output stuff, managing stdout and a log file */

/* put string */
static void dputs(char *buffer, FILE *file)
{
  int bytes;
  if (settings->log_to_stdout) fputs(buffer, stdout);
  if (file != NULL) {
    bytes = fputs(buffer, file);
    if (bytes == EOF) { perror("fputs"); exit(1); }
  }
}

/* put character */
static void dputc(int byte, FILE *file)
{
  if (settings->log_to_stdout) fputc(byte, stdout);
  if (file != NULL) {
    byte = fputc(byte, file);
    if (byte == EOF) { perror("fputc"); exit(1); }
  }
}

/* put a time_t as decimal */
static void dputt(time_t now, FILE *file)
{
  char *buffer = g_strdup_printf("%d", (int) now);
  dputs(buffer, file);
  g_free(buffer);
}

/* flush the output streams */
static void dflush(FILE *file)
{
  if (settings->log_to_stdout) fflush(stdout);
  if (file != NULL) fflush(file);
}

/* convert from call logger date time format to time_t */
time_t to_time(char *p)
{
  /* 01-Dec-01 18:28:11 */
  /* 012345678901234567 */
  struct tm tm;
  time_t here;
  char *hope = "JanFebMarAprMayJunJulAugSepOctNovDec";
  char *chop = g_strndup(p+3,3);
  
  tm.tm_year = 2000 + atoi(p+7) - 1900;
  tm.tm_mon = (strstr(hope, chop) - hope) / 3;
  g_free(chop);
  tm.tm_mday = atoi(p);
  tm.tm_hour = atoi(p+10);
  tm.tm_min = atoi(p+13);
  tm.tm_sec = atoi(p+16);
  tm.tm_isdst = 1;
  here = mktime(&tm);
  /* test harness */
#ifdef TIME_DEBUG
  fprintf(stderr, "%s --> %d %d %d %d %d %d --> %s", p, tm.tm_year, tm.tm_mon+1, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec, ctime(&here));
#endif
  return here;
}

/* convert from call logger date time format to time_t */
time_t to_time_new(char *p)
{
  /* 01-Dec-01 18:28:11 */
  /* d0 d1  d2 t0 t1 t2 */
  struct tm tm;
  char **halves	= g_strsplit(p, " ", 2);
  char **dates	= g_strsplit(halves[0], "-", 3);
  char *day	= dates[0];
  char *month	= dates[1];
  char *year	= dates[2];
  char **times	= g_strsplit(halves[1], "-", 3);
  char *hour    = times[0];
  char *minute	= times[1];
  char *second	= times[2];
  char *hope = "JanFebMarAprMayJunJulAugSepOctNovDec";
  time_t here;
  tm.tm_year = 2000 + atoi(year) - 1900;
  tm.tm_mon = (strstr(hope, month) - hope) / 3;
  tm.tm_mday = atoi(day);
  tm.tm_hour = atoi(hour);
  tm.tm_min = atoi(minute);
  tm.tm_sec = atoi(second);
  tm.tm_isdst = 0;
  here = mktime(&tm);
  /* test harness */
  fprintf(stderr, "%s --> %d %d %d %d %d %d --> %s", p, tm.tm_year, tm.tm_mon+1, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec, ctime(&here));
  g_strfreev(halves);
  g_strfreev(dates);
  g_strfreev(times);
  return here;
}

/* convert from time_t to a file name */
char *to_name(time_t now, char *suffix)
{
  struct tm *tm = localtime(&now);
  char *name = g_strdup_printf("k164-%04d-%02d-%02d%s", tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday, suffix);
  return name;
}

/* convert from time_t to a date */
char *to_date(time_t now)
{
  struct tm *tm = localtime(&now);
  char *name = g_strdup("dd-mmm-yyyy ");
  strftime(name, strlen(name), "%d-%b-%Y", tm);
  return name;
}

#define STATE_READY      0	/* expecting anything			*/
#define STATE_OFF_HOOK   1	/* expecting off hook timestamp		*/
#define STATE_DIALLING   2	/* expecting dial codes			*/
#define STATE_ON_HOOK    3	/* expecting on hook timestamp		*/
#define STATE_RESET	 4	/* expecting rest of reset message	*/

/* negotiation steps */
typedef struct _NegotiationStep NegotiationStep;

struct _NegotiationStep {
  int (*function)(char byte, char *parameter);	/* function to call on byte received	*/
  char *parameter;		/* parameter for the function		*/
};

/* predefined negotiation functions */
int n_put(char byte, char *parameter);	/* put a string with a \r		*/
int n_wait(char byte, char *parameter);	/* wait for a string to be seen		*/
int n_data(char byte, char *parameter);	/* read data until blank line		*/
int n_choice(char byte, char *parameter);	/* wait for Bad Password or Choice?	*/
int n_end(char byte, char *parameter);	/* end the list				*/

/* return values for a negotiation function */
#define N_CONTINUE	0	/* call me for next byte		*/
#define N_NEXT		1	/* move to next item in list		*/
#define N_SUCCESS	2	/* conclude sequence and report success */
#define N_FAILURE	3	/* conclude sequence and report failure */

/* current negotiation step list */
NegotiationStep *cnsl = NULL;
NegotiationStep *cns = NULL;	/* current step				*/
GtkFunction ns_success, ns_failure, ns_save;
gpointer ns_data;

#undef NS_DEBUG

/* start a negotiation step list */
void ns_new(GtkFunction success, GtkFunction failure, gpointer data)
{
  cnsl = cns = g_new0(NegotiationStep, 32);
  ns_success = success;
  ns_failure = failure;
  ns_data = data;
}

void ns_add(NegotiationStep *cns, int (*function)(char byte, char *parameter), char *parameter)
{
  cns->function = function;
  cns->parameter = parameter;
}

/* delete the list */
void ns_free()
{
  g_free(cnsl);
  cnsl = cns = NULL;
}

/* called by reader for each byte read */
void ns_read(char byte)
{
  int result = N_NEXT;

#ifdef NS_DEBUG
  fprintf(stderr, "ns_read: step %d, byte %02x %c ", (int) (cns - cnsl), byte, (byte > 32) ? byte : 32);
#endif

  while (1) {
#ifdef NS_DEBUG
    if (cns->function == n_put) fprintf(stderr, "n_put '%s' ", cns->parameter);
    if (cns->function == n_wait) fprintf(stderr, "n_wait '%s' ", cns->parameter);
    if (cns->function == n_data) fprintf(stderr, "n_data ");
    if (cns->function == n_choice) fprintf(stderr, "n_choice ");
    if (cns->function == n_end) fprintf(stderr, "n_end ");
#endif

    result = (*cns->function) (byte, cns->parameter);

    switch (result) {
    case N_CONTINUE:	/* call me for next byte		*/
#ifdef NS_DEBUG
      fprintf(stderr, "N_CONTINUE\n");
#endif
      return;
      break;
    case N_SUCCESS:	/* conclude sequence and report success */
      ns_free();
#ifdef NS_DEBUG
      fprintf(stderr, "N_SUCCESS\n");
#endif
      if (ns_success != NULL) (*ns_success) (ns_data);
      menu_device_set_sensitive(1);
      return;
      break;
    case N_FAILURE:	/* conclude sequence and report failure */
      ns_free();
#ifdef NS_DEBUG
      fprintf(stderr, "N_FAILURE\n");
#endif
      if (ns_failure != NULL) (*ns_failure) (ns_data);
      menu_device_set_sensitive(1);
      return;
      break;
    case N_NEXT:
#ifdef NS_DEBUG
      fprintf(stderr, "N_NEXT\n");
#endif
      break;
    }

    cns++;
  }
}

/* execute the first step in a negotiation step list */
void ns_execute()
{
  cns = cnsl;
  ns_read(0);
  menu_device_set_sensitive(0);
}

void ns_timeout()
{
  ns_free();
  menu_device_set_sensitive(1);
}

/* negotiate to get the menu up */
void ns_menu(char *password)
{  
  ns_add(cns++, n_put, "");
  ns_add(cns++, n_wait, "Password? ");
  ns_add(cns++, n_put, password);
  ns_add(cns++, n_choice, NULL);
}

/* exit */
void ns_exit()
{
  ns_add(cns++, n_put, "0");
  ns_add(cns++, n_wait, "Ready\r\n");
  ns_add(cns++, n_end, NULL);
}

/* test connection */
void ns_test(GtkFunction success, GtkFunction failure, gpointer data)
{
  ns_new(success, failure, data);
  ns_add(cns++, n_put, "0");
  ns_add(cns++, n_wait, "Password? ");
  ns_add(cns++, n_put, "");
  ns_add(cns++, n_wait, "Ready\r\n");
  ns_add(cns++, n_end, NULL);
  ns_execute(0);
}

/* download data */
void ns_download(GtkFunction success, GtkFunction failure, GtkFunction save, gpointer data, char *password)
{
  ns_new(success, failure, data);
  ns_save = save;
  ns_menu(password);
  ns_add(cns++, n_put, "1");
  ns_add(cns++, n_wait, "\r\n");
  ns_add(cns++, n_data, NULL);
  ns_add(cns++, n_choice, NULL);
  ns_exit();
  ns_execute(0);
}

/* clear data */
void ns_clear(GtkFunction success, GtkFunction failure, gpointer data, char *password)
{
  ns_new(success, failure, data);
  ns_menu(password);
  ns_add(cns++, n_put, "2");
  /* ?? */
  ns_add(cns++, n_choice, NULL);
  ns_exit();
  ns_execute(0);
}

/* set date and time */
void ns_set_date_and_time(GtkFunction success, GtkFunction failure, gpointer data, char *password, char *date, char *time)
{
  ns_new(success, failure, data);
  ns_menu(password);
  ns_add(cns++, n_put, "3");
  ns_add(cns++, n_wait, "(mm/dd/yy)? ");
  ns_add(cns++, n_put, date);
  ns_add(cns++, n_wait, "Choice? ");
  ns_add(cns++, n_put, "4");
  ns_add(cns++, n_wait, "(hh:mm:ss)? ");
  ns_add(cns++, n_put, time);
  ns_add(cns++, n_choice, NULL);
  ns_exit();
  ns_execute(0);
}

/* set password */
void ns_change_password(GtkFunction success, GtkFunction failure, gpointer data, char *old, char *new)
{
  ns_new(success, failure, data);
  ns_menu(old);
  ns_add(cns++, n_put, "5");
  ns_add(cns++, n_wait, "New Password? ");
  ns_add(cns++, n_put, new);
  ns_add(cns++, n_choice, NULL);
  ns_exit();
  ns_execute(0);
}

#define MAXN 128
char n_buffer[MAXN], *n_here;

/* put a string with a \r		*/
int n_put(char byte, char *parameter)
{
  int bytes;
  char *cr = "\r";

  bytes = write(fd, parameter, strlen(parameter));
  if (bytes < 0) { perror("write"); exit(1); }
  bytes = write(fd, cr, strlen(cr));
  if (bytes < 0) { perror("write"); exit(1); }

  n_here = n_buffer;
  return N_NEXT;
}

/* wait for a string to be seen		*/
int n_wait(char byte, char *parameter)
{
  *n_here++ = byte;
  *n_here = '\0';
  if (!strncmp(n_here - strlen(parameter), parameter, strlen(parameter))) {
    n_here = n_buffer;
    return N_NEXT;
  } else {
    return N_CONTINUE;
  }
}

FILE *n_data_output;

/* read data until blank line		*/
int n_data(char byte, char *parameter)
{
  /* ?? can download more than once? this is not reset */
  static int previous = -1;

  if (byte == '\r') return N_CONTINUE;

  *n_here++ = byte;
  *n_here = '\0';

  if (byte == '\n') {
    (*ns_save) (ns_data);
    n_here = n_buffer;
    if ((previous == 1) && (strlen(n_buffer) == 1)) return N_NEXT;
    previous = strlen(n_buffer);
  }
  return N_CONTINUE;
}

/* wait for Bad Password or Choice?	*/
int n_choice(char byte, char *ignored)
{
  char *parameter;

  *n_here++ = byte;
  *n_here = '\0';

  parameter = "Bad password";
  if (!strncmp(n_here - strlen(parameter), parameter, strlen(parameter))) {
    return N_FAILURE;
  }

  parameter = "Choice? ";
  if (!strncmp(n_here - strlen(parameter), parameter, strlen(parameter))) {
    return N_NEXT;
  }

  return N_CONTINUE;
}

/* end the list				*/
int n_end(char byte, char *parameter)
{
  return N_SUCCESS;
}

FILE *output = NULL;

/* process data arriving on the serial port */
static void reader (gpointer ignored, gint source, GdkInputCondition condition)
{
  static time_t h0, h1, tx, tn;
  static int byte, n;
  static char *p, time_0[MAXTIME], data[MAXDATA], time_1[MAXTIME], *name;

  /* read the data */
  int bytes = read(fd, &byte, 1);
  if (bytes == 0) return;
  if (bytes < 0) {
    perror("read");
    return;
  }

  /* negotiation step parser intercept, too much caffeine */
  if (cns != NULL) {
    ns_read(byte);
    return;
  }

  /* normal processing of spontaneous data from logger */
  switch (state) {

  case STATE_READY:
    /* waiting for next message */

    switch (byte) {
    case '\r': break;
    case '\n':
      state = STATE_OFF_HOOK;
      t0 = time(NULL);
      p = time_0;
      n = 0;
      menu_device_set_sensitive(0);
      break;
    case 0:
      entry_generic("--:--:--", "NUL", "--:--", "--:--:--");
      statusbar_timed("Received a NUL, was logger power cycled?", 2000);
      break;
    case ',':
      entry_generic("--:--:--", "SYNC", "--:--", "--:--:--");
      statusbar_timed("Received a comma, was GUI started after call?", 2000);
      break;
    default:
      statusbar_timed("Unexpected data from logger, see stderr", 2000);
      fprintf(stderr, "k164d: state 0 unexpected character %02x\n", byte);
      break;
    }
    break;


  case STATE_OFF_HOOK:
    /* read the date and time stamp dd-mmm-yy hh:mm:ss */
    switch (byte) {

    case '\r':
      state = STATE_READY;
      menu_device_set_sensitive(1);
      break;

    case 'T': /* detect a reset sequence */
      *p = '\0';
      state = STATE_RESET;
      n = 0;
      statusbar_timed("Logger reset", 2000);
      fprintf(stderr, "k164d: detected reset sequence, %s\n", time_0);
      break;
      
    case ',':
      *p = '\0';
      state = STATE_DIALLING;
      h0 = to_time(time_0);
      
      /* create a file name */
      if (settings->log_to_file) {
	name = to_name(h0, ".log");
	output = fopen(name, "a");
	if (output == NULL) { perror(name); exit(1); }
      } else {
	output = NULL;
      }

      /* begin the output */
      dputs(time_0, output);
      dputc(',', output);
      dputt(h0, output);
      dputc(',', output);
      dflush(output);
      
      entry_begin(time_0);
      statusbar_timed("Off-hook", 2000);
      
      p = data; n = 0; tx = time(NULL);
      break;
    default:
      if (n++ < MAXTIME) *p++ = byte;
      break;
    }
    break;

  case STATE_DIALLING:
    switch (byte) {
    case ',':
      state = STATE_ON_HOOK;
      *p = '\0';
      dputc(',', output);
      dputs(data, output);
      dputc(',', output);
      t1 = time(NULL); p = time_1; n = 0;
      break;
    default:
      tn = time(NULL);
      if ((tn - tx) > 1) dputc('=', output);
      tx = tn;
      if (n++ < MAXDATA) *p++ = byte;
      entry_append(byte);
      dputc(byte, output);
      dflush(output);
      statusbar_timed("Dialling", 500);
      break;
    }
    break;

  case STATE_ON_HOOK:
    /* read the date and time stamp dd-mmm-yy hh:mm:ss */
    switch (byte) {
    case '\n':
      state = STATE_READY;
      *p = '\0';
      h1 = to_time(time_1);
      
      dputs(time_1, output);
      dputc(',', output);
      dputt(h1, output);
      
      /* emit duration calculated from input data */
      p = g_strdup_printf(",%d", (int) (h1 - h0));
      dputs(p, output);
      g_free(p);
      
      /* emit error calculated from our timing */
      p = g_strdup_printf(",%d\n", (int) (h1 - h0) - (int) (t1 - t0));
      dputs(p, output);
      g_free(p);
      
      dflush(output);
      if (output != NULL) fclose(output);
      output = NULL;
      g_free(name);

      entry_close((int) (h1 - h0), time_1);
      statusbar_timed("On hook", 5000);
      menu_device_set_sensitive(1);
      break;
    case '\r': break;
    default:
      if (n++ < MAXTIME) *p++ = byte;
      break;
    }
    break;

  case STATE_RESET:
    /* \r\nTelephone Call Logger V1.9 (16K)\r\n\r\nReady\r\n */
    /*      ^                                                */
    switch (byte) {
    case 'R': n++; break;
    case 'y': n++; break;
    case '\n':
      if (n > 1) { 
	state = STATE_READY; 
	entry_generic("--:--:--", "RESET", "--:--", "--:--:--");
	menu_device_set_sensitive(1);
      }
      break;
    }
    break;
  }
}

/* set a file descriptor non-blocking */
static void unblock (int fd)
{
    int flags;
    flags = O_NONBLOCK | fcntl(fd, F_GETFL);
    fcntl(fd, F_SETFL, flags);
}

/* start listening */
static void initialise ()
{
  struct termios termios;
  FILE *input;

  /* allow device to default */
  if (device == NULL) device = "/dev/ttyS1";

  /* open the serial port */
  input = fopen(device, "r+" );
  if (input == NULL) {
    char *text;

    text = g_strdup_printf("fopen: %s: %s\n\n"

			   "Could not open the %s tty device to accept data\n"
			   "from the microcontroller.  You may need to use\n"
			   "the command line option to specify a device.\n\n"
			   "For example; k164 -f /dev/ttyS2\n", device,
			   strerror(errno), device);
    message("k164: fopen failure", text);
    g_free(text);
    return;
  }

  /* get and set the serial port characteristics */
  fd = fileno(input);
  if (tcgetattr(fd, &termios) < 0) {
    char *text;

    text = g_strdup_printf("tcgetattr: %s: %s\n\n"

			   "Could not obtain tty attributes of %s.\n\n"
			   "This likely means the device you asked me to\n"
			   "use is not a tty, or the default ttyS1 is not\n"
			   "a valid serial port on your machine.\n\n"
			   "You need to provide a different device on the\n"
			   "command line as an option; for example\n\n"
			   "# k164 -f /dev/ttyS0\n\n",
			   device, strerror(errno), device);

    message("k164: tcgetattr failure", text);
    g_free(text);
    return;
  }

  if (cfsetospeed(&termios, B9600) < 0) {
    perror("cfsetospeed");
    exit(1);
  }

  if (cfsetispeed(&termios, B9600) < 0) {
    perror("cfsetispeed");
    exit(1);
  }

  cfmakeraw(&termios);

  if (tcsetattr(fd, TCSANOW, &termios) < 0) {
    perror("tcsetattr");
    exit(1);
  }

  /* enable non-blocking i/o */
  unblock(fd);

  /* inform gtk that we want to be told when data arrives on fd */
  in = gtk_input_add_full(fd, GDK_INPUT_READ, reader, NULL, NULL, NULL);

  settings = g_new0(Settings, 1);
  settings->log_to_file++;
  settings->log_to_stdout++;
  
  menu_hide("call");
  menu_hide("edit");
  menu_hide("delete");
  menu_hide("comments");
  menu_hide("flag");

  statusbar_timed("Ready", 2000);
}

/* accept the device argument from main() */
void k164_set_device(char *argv)
{
  device = argv;
}


gboolean
on_k164_window_delete_event            (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  gtk_main_quit();
  return FALSE;
}


void
on_k164_window_show                    (GtkWidget       *widget,
                                        gpointer         user_data)
{
  clist = GTK_CLIST(lookup_widget(widget, "clist"));
  gtk_clist_set_column_auto_resize(clist, 0, 1);
  gtk_clist_set_column_auto_resize(clist, 1, 1);
  gtk_clist_set_column_auto_resize(clist, 2, 1);
  gtk_clist_set_column_auto_resize(clist, 3, 1);

  initialise();
}


void
on_quit_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gtk_main_quit();
}


void
on_about_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  message ("About k164 Telephone Call Logger",
"k164 version 1.0, Copyright (C) 2001 James Cameron, "
"<quozl@us.netrek.org>\n"
"k164 comes with ABSOLUTELY NO WARRANTY; for details see source.\n"
"This is free software, and you are welcome to redistribute it "
"under certain conditions;\nsee source for details.\n"
"\n"
"k164 is a program for capturing data from a kit supplied by http://kitsrus.com/\n"
"\n"
"Problem reports, suggestions, and patches are welcome.\n");
}


void
on_clist_click_column                  (GtkCList        *clist,
                                        gint             column,
                                        gpointer         user_data)
{
  static gint prior = -1;
  GtkSortType order;
  char *name, *text;

  /* set sort order, was it a second click on same column? */
  if (prior == column) {
    order = GTK_SORT_DESCENDING;
    prior = -1;
  } else {
    order = GTK_SORT_ASCENDING;
    prior = column;
  }

  /* get the column name */
  switch (column) {
  case 0: name = "off hook time"; break;
  case 1: name = "number dialled"; break;
  case 2: name = "elapsed time"; break;
  case 3: name = "on hook time"; break;
  default: name = "unknown"; break;
  }

  /* generate the statusbar text */
  text = g_strdup_printf("Sort by %s%s", 
			 order == GTK_SORT_DESCENDING ? "descending " : "",
			 name);
  statusbar_timed(text, 2000);
  g_free(text);

  /* sort the list */
  gtk_clist_set_sort_type(clist, order);
  gtk_clist_set_sort_column(clist, column);
  gtk_clist_sort(clist);
}

void
on_password_entry_activate             (GtkEditable     *editable,
                                        gpointer         user_data)
{

}

GtkWidget *password_new(char *title, char *label)
{
  GtkWidget *k164_password = create_k164_password();
  gtk_window_set_title(GTK_WINDOW(k164_password), title);
  gtk_label_set_text(GTK_LABEL(lookup_widget(k164_password, "label")), 
		     label);
  gtk_widget_show(k164_password);
  return k164_password;
}

void password_set_activate(GtkWidget *password, GtkSignalFunc activate, 
			   gpointer user_data)
{
  gtk_signal_connect(GTK_OBJECT(lookup_widget(password, "entry")), 
		      "activate", activate, user_data);
}

void password_set_delete(GtkWidget *password, GtkSignalFunc delete, 
			 gpointer user_data)
{
  gtk_signal_connect(GTK_OBJECT(password), "delete_event",
		     delete, user_data);
}

GtkWidget *password(char *title, char *label, GtkSignalFunc activate, 
			   gpointer user_data)
{
  GtkWidget *pw = password_new(title, label);
  password_set_activate(pw, activate, user_data);
  return pw;
}

void password_cancel(GtkEditable *editable)
{
  gtk_widget_destroy(gtk_widget_get_toplevel(GTK_WIDGET(editable)));
}

gchar *password_get(GtkEditable *editable)
{
  return g_strdup(gtk_entry_get_text(GTK_ENTRY(editable)));
}

int password_valid(char *pw)
{
  int valid = 1;
  gchar *text = g_strdup("");

  /* validity tests, k164 documentation, page 3, column 2, list item 3 */
  if (strlen(pw) > 8) {
    gchar *new = g_strconcat(text, "\n- greater than eight characters.");
    g_free(text); text = new;
    valid = 0;
  }
  if (strlen(pw) < 4) {
    gchar *new = g_strconcat(text, "\n- less than four characters.");
    g_free(text); text = new;
    valid = 0;
  }

  if (!valid) {
    gchar *new = g_strdup_printf("The password you typed is not valid:%s", text);
    message ("k164 change password - invalid password", new);
    g_free(new);
  }
  g_free(text);
  return valid;
}






/* download stuff */

typedef struct _Download Download;

struct	_Download
{
  gchar *password;
  guint timeout;
  FILE *file;
  char *tmp;
  char *end;
  int count;
  time_t now, oldest, newest;
};

void download_free(Download *d)
{
  if (d != NULL) {
    g_free(d->password);
    g_free(d);
  }
}

/* ?? failure is overloaded, it means bad password and timeout, split
   the functions so as to report them correctly, and also to call
   ns_timeout() safely, and allow the device menu to become sensitive
   again. */

static int on_download_failure(Download *d)
{
  gtk_timeout_remove(d->timeout);
  gtk_clist_thaw(clist);
  fclose(d->file);
  unlink(d->tmp);
  message ("k164 download failure", "Bad password");
  download_free(d);
  return FALSE;
}

static int on_download_success(Download *d)
{
  gtk_progress_bar_update(progressbar, 0.0);
  gtk_timeout_remove(d->timeout);
  gtk_clist_thaw(clist);
  if (fclose(d->file) != 0) {
    char *text;
    text = g_strdup_printf("k164 download: %s: %s\n\n"
			   "Could not close the file %s to save "
			   "the download data.\n",
			   d->tmp, strerror(errno), d->tmp);
    message("k164 download: fclose: failed", text);
    g_free(text);
  } else {
    if (rename(d->tmp, d->end) != 0) {
      char *text;
      text = g_strdup_printf("k164 download: %s: %s\n\n"
			     "Could not move the downloaded file %s\n"
			     "to the final name of %s",
			     d->end, strerror(errno), d->tmp, d->end);
      message("k164 download: rename: failed", text);
      g_free(text);
    } else {
      char *oldest, *newest, *text;
      oldest = to_date(d->oldest);
      newest = to_date(d->newest);
      text = g_strdup_printf("Downloaded to\n"
			     "%s\n"
			     "%d call records from %s to %s",
			     d->end, d->count, oldest, newest);
      message("k164 download success", text);
      statusbar_timed("Download completed", 2500);
      g_free(oldest);
      g_free(newest);
      g_free(text);
    }
  }

  download_free(d);
  return FALSE;
}

static int on_download_save(Download *d)
{
  /* ?? scope, why do we know about n_buffer? */
  fprintf(d->file, "%s", n_buffer);
  /* ?? what if write() fails? */

  if (strlen(n_buffer) > 10) {
    gchar **array = g_strsplit(n_buffer, ",", 3);
    char *time_0 = array[0];
    char *time_1 = array[1];
    char *dial = g_strchomp(array[2]);
    time_t t0 = to_time(time_0);
    time_t t1 = to_time(time_1);
    char *elapsed = to_elapsed((int)(t1 - t0));

    if (settings->privacy) dial = g_strndup(dial, PRIVACY);

    /* put the line into the list */
    entry_generic(time_0, dial, elapsed, time_1);
    if (settings->privacy) g_free(dial);
    g_free(elapsed);

    /* keep track of call record range */
    if (d->count++ == 1) {
      d->newest = t1;
      d->oldest = t0;
    } else {
      if (t1 > d->newest) { d->newest = t1; }
      if (t0 < d->oldest) { d->oldest = t0; }

      /* running commentary */
      if ((d->count % 10) == 0) {
	char *text = g_strdup_printf("%6d call records", d->count);
	statusbar_timed(text, 500);
	g_free(text);
      }

      /* progress bar */
      {
	float progress = (float) (t1 - d->oldest) / (d->now - d->oldest);
	if (progress >= 0 && progress <= 1.0)
	  gtk_progress_bar_update(progressbar, progress);
      }
    }
  
    /* cleanup */
    g_strfreev(array);
  }

  /* restart the timeout */
  gtk_timeout_remove(d->timeout);
  d->timeout = gtk_timeout_add(500, (GtkFunction) on_download_failure, d);
  return FALSE;
}

void
on_download                            (GtkEditable     *editable,
                                        gpointer         user_data)
{
  time_t now = time(NULL);
  char *end = to_name(now, ".dump");
  char *tmp = g_strdup_printf("%s.tmp", end);

  n_data_output = fopen(tmp, "w");
  if (n_data_output == NULL) {
    char *text;
    text = g_strdup_printf("k164 download: %s: %s\n\n"
			   "Could not open the file %s to save "
			   "the download data.\n",
			   tmp, strerror(errno), tmp);
    message("k164 download: open: failed", text);
    g_free(text);
  } else {
    Download *d = g_new0(Download, 1);

    d->timeout = gtk_timeout_add(1000, (GtkFunction) on_download_failure, d);
    d->password = password_get(editable);
    d->file = n_data_output;
    d->tmp = g_strdup(tmp);
    d->end = g_strdup(end);
    d->count = 0;
    d->now = now;

    gtk_clist_clear(clist);
    gtk_clist_freeze(clist);
    gtk_progress_bar_update(progressbar, 0.0);

    ns_download((GtkFunction) on_download_success, 
		(GtkFunction) on_download_failure, 
		(GtkFunction) on_download_save, 
		d, d->password);
  }
  password_cancel(editable);
}


void
on_download_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  password("Download", "Password : ", on_download, NULL);
  statusbar_timed("Download", 2000);
}


void
on_download_and_clear                  (GtkEditable     *editable,
                                        gpointer         user_data)
{
  password_cancel(editable);
  message ("k164 download and clear - unimplemented", 
	   "Apologies, this function is not yet enabled,\n"
	   "because the author wants to wait for his logger\n"
	   "to fill up enough to warrant it.");
  /*

   */
}

void
on_download_and_clear_activate         (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  password("Download and Clear", "Password : ", on_download_and_clear, NULL);
  statusbar_timed("Download and Clear", 2000);
}




/* set date and time stuff */

typedef struct _SetDateAndTime SetDateAndTime;

struct	_SetDateAndTime
{
  gchar	*password, *date, *time;
  guint timeout;
};

void set_date_and_time_free(SetDateAndTime *sdat)
{
  if (sdat != NULL) {
    g_free(sdat->password);
    g_free(sdat->date);
    g_free(sdat->time);
    g_free(sdat);
  }
}

static int on_set_date_and_time_failure(SetDateAndTime *sdat)
{
  gtk_timeout_remove(sdat->timeout);
  message ("k164 set date and time failure", "Bad password");
  set_date_and_time_free(sdat);
  return FALSE;
}

static int on_set_date_and_time_success(SetDateAndTime *sdat)
{
  gtk_timeout_remove(sdat->timeout);
  message ("k164 set date and time success",
	   "Date and time set successfully.");
  set_date_and_time_free(sdat);
  return FALSE;
}

void
on_set_date_and_time                   (GtkEditable     *editable,
                                        gpointer         user_data)
{
  SetDateAndTime *sdat = g_new0(SetDateAndTime, 1);
  time_t now = time(NULL);
  struct tm *tm = localtime(&now);

  sdat->timeout = gtk_timeout_add(2000, 
				  (GtkFunction) on_set_date_and_time_failure, 
				  sdat);

  sdat->password = password_get(editable);
  password_cancel(editable);

  sdat->date = g_strdup_printf("%02d/%02d/%02d", 
			       tm->tm_mon + 1, tm->tm_mday, 
			       tm->tm_year + 1900 - 2000);

  sdat->time = g_strdup_printf("%02d:%02d:%02d", 
			       tm->tm_hour, tm->tm_min, tm->tm_sec);

  ns_set_date_and_time((GtkFunction) on_set_date_and_time_success, 
		       (GtkFunction) on_set_date_and_time_failure, 
		       sdat, sdat->password, sdat->date, sdat->time);
}

void
on_set_date_and_time_activate          (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  password("Set Date and Time", "Password : ", on_set_date_and_time, NULL);
  statusbar_timed("Set Date and Time", 2000);
}




/* change password stuff */

typedef struct _ChangePassword ChangePassword;

struct	_ChangePassword
{
  gchar	*old;		/* old password			*/
  gchar	*new;		/* new password			*/
  gchar	*confirm;	/* confirmed new password	*/
  guint timeout;
};

void change_password_free(ChangePassword *cp)
{
  if (cp != NULL) {
    if (cp->old != NULL) g_free(cp->old);
    if (cp->new != NULL) g_free(cp->new);
    if (cp->confirm != NULL) g_free(cp->confirm);
    g_free(cp);
  }
}

gboolean
on_change_password_delete              (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        ChangePassword  *cp)
{
  change_password_free(cp);
  return FALSE;
}

static int on_change_password_failure(ChangePassword *cp)
{
  gtk_timeout_remove(cp->timeout);
  message ("k164 change password failure", "Bad password");
  change_password_free(cp);
  return FALSE;
}

static int on_change_password_success(ChangePassword *cp)
{
  gtk_timeout_remove(cp->timeout);
  message ("k164 change password success",
	   "Password changed successfully.");
  change_password_free(cp);
  return FALSE;
}

void
on_change_password_confirm             (GtkEditable     *editable,
                                        ChangePassword  *cp)
{
  cp->confirm = password_get(editable);
  password_cancel(editable);
  if (password_valid(cp->confirm)) {
    if (!strcmp(cp->new, cp->confirm)) {
      cp->timeout = gtk_timeout_add(2000, 
				    (GtkFunction) on_change_password_failure, 
				    cp);
      
      ns_change_password((GtkFunction) on_change_password_success, 
			 (GtkFunction) on_change_password_failure, 
			 cp, cp->old, cp->new);
    } else {
      message("k164 change password - not the same", 
	       "Passwords did not match.\n"
	       "\n"
	       "What you entered as the new password did not\n"
	       "not match what you entered as a confirmation.");
      statusbar_timed("Change Password - Confirmation Failure", 2000);
      change_password_free(cp);
    }
  } else {
    change_password_free(cp);
  }
}


void
on_change_password_new                 (GtkEditable     *editable,
                                        ChangePassword  *cp)
{
  GtkWidget *pw;
  cp->new = password_get(editable);
  password_cancel(editable);
  if (password_valid(cp->new)) {
    pw = password("Change Password", "Confirmation : ", 
		  on_change_password_confirm, cp);
    password_set_delete(pw, GTK_SIGNAL_FUNC(on_change_password_delete), cp);
    statusbar_timed("Change Password - Confirm Password", 2000);
  } else {
    change_password_free(cp);
  }
}

void
on_change_password_old                 (GtkEditable     *editable,
                                        ChangePassword  *cp)
{
  GtkWidget *pw;
  cp->old = password_get(editable);
  password_cancel(editable);
  if (password_valid(cp->old)) {
    pw = password("Change Password", "New Password : ", 
		  on_change_password_new, cp);
    password_set_delete(pw, GTK_SIGNAL_FUNC(on_change_password_delete), cp);
    statusbar_timed("Change Password - New Password", 2000);
  } else {
    change_password_free(cp);
  }
}


void
on_change_password_activate            (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  ChangePassword *cp = g_new0(ChangePassword, 1);
  GtkWidget *pw = password("Change Password", "Old Password : ", 
			   on_change_password_old, cp);
  password_set_delete(pw, GTK_SIGNAL_FUNC(on_change_password_delete), cp);
  statusbar_timed("Change Password - Old Password", 2000);
}


static int on_test_failure(guint *timeout)
{
  g_free(timeout);
  message ("k164 test failure", 
		 "Failure.\n\n"
		 "Incorrect or inadequate response, some possible causes:\n"
		 "(a) telephone call logger has no power,\n"
		 "(b) cable to logger is disconnected or faulty,\n"
		 "(c) serial port is faulty,\n"
		 "(d) insufficient privilege to access serial port,\n"
		 "(e) logger is at menu prompt and this program can't tell,\n"
		 "(f) this program has a design fault.\n");
  statusbar_timed("Test Failure", 3000);
  return FALSE;
}

static int on_test_success(guint *timeout)
{
  gtk_timeout_remove(*timeout);
  g_free(timeout);
  message ("k164 test success", 
	   "Success.\n\n"
	   "Response from telephone call logger was as expected.");
  statusbar_timed("Test Success", 3000);
  return FALSE;
}

void
on_test_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  guint *timeout = g_new(guint, 1);
  *timeout = gtk_timeout_add(2000, (GtkFunction) on_test_failure, timeout);
  ns_test((GtkFunction) on_test_success, (GtkFunction) on_test_failure, timeout);
  statusbar_timed("Testing", 2000);
}


void
on_edit_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


void
on_delete_activate                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


void
on_log_to_file_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkCheckMenuItem *us = GTK_CHECK_MENU_ITEM(menuitem);
  settings->log_to_file = us->active;
}


void
on_log_to_stdout_activate              (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkCheckMenuItem *us = GTK_CHECK_MENU_ITEM(menuitem);
  settings->log_to_stdout = us->active;
}


void
on_comments_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkCheckMenuItem *us = GTK_CHECK_MENU_ITEM(menuitem);
  if (us->active) {
  }
}


void
on_flag_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkCheckMenuItem *us = GTK_CHECK_MENU_ITEM(menuitem);
  if (us->active) {
  }
}


void
on_privacy_activate                    (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkCheckMenuItem *us = GTK_CHECK_MENU_ITEM(menuitem);
  settings->privacy = us->active;
}



